# Copyright (c) 2003, 2005 Brian Katzung, Kappa Computer Solutions, LLC
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# For more information, contact:
#
#	Kappa Computer Solutions, LLC
#	http://www.kappacs.com/
#
# 2005-07-13	Constant and initial values can now be passed either
#		as a list or as a hash ref.
# 2005-03-05	Initial and constant field value support added by
#		Brian Katzung.
# 2003-10-29	Created by Brian Katzung.
#

package Text::FF::Layout;

use strict;

# Index of layouts by name
my %index;

#======================================================================
# Functions

#----------------------------------------------------------------------
# Add constant field values (instance method)
# Returns the instance.

sub add_constant
{
	my ($self, @values) = @_;

	$self->{'constant'} = { %{$self->{'constant'}},
	  (@values && ref($values[0]))? %{$values[0]}: @values };

	return $self;
}

#----------------------------------------------------------------------
# Add initial field values (instance method)
# Returns the instance.

sub add_initial
{
	my ($self, @values) = @_;

	$self->{'initial'} = { %{$self->{'initial'}},
	  (@values && ref($values[0]))? %{$values[0]}: @values };

	return $self;
}

#----------------------------------------------------------------------
# Clone a layout (instance method)
# Returns the instance.

sub clone
{
	my ($self, $from) = @_;

	$from = $self->get_by_name($from)
	  if ($from && !ref($from));

	%$self = ( %$from, 'name' => $self->{'name'} ) if ($from);

	return $self;
}

#----------------------------------------------------------------------
# Return hash of constant field values (instance method)

sub get_constant
{
	my ($self) = @_;

	return $self->{'constant'};
}

#----------------------------------------------------------------------
# Return a layout by name (class or instance method)

sub get_by_name
{
	my ($self, $name) = @_;

	return $index{$name};
}

#----------------------------------------------------------------------
# Return the field specifications (instance method)

sub get_field_specs
{
	my ($self) = @_;

	return $self->{'field_spec'};
}

#----------------------------------------------------------------------
# Return hash of initial field values (instance method)

sub get_initial
{
	my ($self) = @_;

	return $self->{'initial'};
}

#----------------------------------------------------------------------
# Return the maximum record size (instance method)

sub max_size
{
	my ($self) = @_;
	my $max = 0;

	foreach my $spec (values(%{$self->{'field_spec'}}))
	{
		my ($start, $len) = @$spec;

		$len += $start;
		$max = $len if ($len > $max);
	}

	return $max;
}

#----------------------------------------------------------------------
# Create a new layout (class method)
# Returns the instance.

sub new
{
	my ($class, %opt) = @_;
	my $self = bless { 'constant' => {}, 'initial' => {} }, $class;

	$self->clone($opt{'clone'}) if (defined($opt{'clone'}));

	($index{$opt{'name'}} = $self)->{'name'} = $opt{'name'}
	  if (defined($opt{'name'}));
	$self->{'default_type'} = $opt{'default_type'} || 'l';
	$self->add_constant($opt{'constant'}) if ($opt{'constant'});
	$self->add_initial($opt{'initial'}) if ($opt{'initial'});

	$self->parse_layout($opt{'layout'}, $opt{'type_map'})
	  if ($opt{'layout'});

	$self->pad(defined($opt{'pad'})? $opt{'pad'}: 1);

	return $self;
}

#----------------------------------------------------------------------
# Set/get padding flag (instance method)

sub pad
{
	my ($self, $pad) = @_;

	$self->{'pad'} = $pad if (defined($pad));

	return $self->{'pad'};
}

#----------------------------------------------------------------------
# Parse field specifications (instance method)
# Returns the instance.

sub parse_layout
{
	my ($self, $layout, $type_map) = @_;
	my @layout;
	my $next_char = 0;

	@layout = ref($layout)? @$layout: split(/\s+/s, $layout);

	while (@layout > 1)
	{
		my $name = shift(@layout);
		my $next = shift(@layout);
		my $spec;
		my $type;

		if ($next =~ /^\d+(?:[+-]\d+)?$/)
		{
			# The 2nd word is a position or size specification.
			# Use the default type.
			$spec = $next;
			$type = $self->{'default_type'};
		}
		else
		{
			# The 2nd word is a type name or type.
			$type = ($type_map && $type_map->{$next})?
			  $type_map->{$next}: $next;

			# Replace underscores in the type with spaces.
			$type =~ s/_/ /g;

			if ($type =~ /^(\d+)\s*(.*)/)
			{
				# The type contains the size.
				($spec, $type) = ($1, $2);
				$type ||= $self->{'default_type'};
			}
			else
			{
				# Get the position or size from the 3rd word.
				$spec = shift(@layout);
			}
		}

		if ($spec =~ /(\d+)\+(\d+)/)	# start + size
		{
			$self->{'field_spec'}{$name} =
			  [ $1 - 1, $2 + 0, $type ];
			$next_char += $2;
		}
		elsif ($spec =~ /(\d+)-(\d+)/)	# start - stop
		{
			$self->{'field_spec'}{$name} =
			  [ $1 - 1, $2 - $1, $type ];
			$next_char = $2;
		}
		else				# size
		{
			$self->{'field_spec'}{$name} =
			  [ $next_char, $spec + 0, $type ];
			$next_char += $spec;
		}
	}

	return $self;
}

#----------------------------------------------------------------------
# Set constant field values (instance method)
# Returns the instance.

sub set_constant
{
	my ($self, @values) = @_;

	$self->{'constant'} = {};
	$self->add_constant(@values);
	return $self;
}

#----------------------------------------------------------------------
# Set initial field values (instance method)
# Returns the instance.

sub set_initial
{
	my ($self, @values) = @_;

	$self->{'initial'} = {};
	$self->add_initial(@values);
	return $self;
}

1;

__END__

=head1 NAME

Text::FF::Layout - Fixed-field text record layout

=head1 SYNOPSIS

  use Text::FF::Layout;

  # Create a new layout instance
  $layout = Text::FF::Layout->new(I<options>);

  # Retrieve an existing layout by name
  $layout = Text::FF::Layout->get_by_name($name);

  # Parse a layout description
  $layout->parse_layout($description, $map);

  # Clone an existing layout (everything except the name)
  # by name or instance
  $layout->clone($another_layout);

  # Set/add constant field values (always applied)
  $layout->set_constant(%list);    # (or { list })
  $layout->add_constant(%list);    # (or { list })

  # Set/add initial field values (applied whenever layout is set)
  $layout->set_initial(%list);    # (or { list })
  $layout->add_initial(%list);    # (or { list })

  # Specify whether record will be minimum necessary or
  # maximum possible length.
  $layout->pad($bool);

=head1 DESCRIPTION

Instances of this class described fixed-field text record layouts.
Instances of Text::FF::Record use these layouts to access the fields.

=head2 new Options

=over 4

=item 'clone' => $another_layout

This is equivalent to C<$layout = Text::FF::Layout-E<gt>new()-E<gt>clone($another_layout);>.

=item 'constant' => { %list }

This is equivalent to C<$layout = Text::FF::Layout-E<gt>new()-E<gt>add_constant(%list);>.

=item 'default_type' => $type

This sets the default field type. The default default type is 'l'.
See the section on layout descriptions.

=item 'initial' => { %list }

This is equivalent to C<$layout = Text::FF::Layout-E<gt>new()-E<gt>add_initial(%list);>.

=item 'layout' => $description, 'type_map' => $map

This is equivalent to C<$layout = Text::FF::Layout-E<gt>new()-E<gt>parse_layout($description, $map);>.

=item 'name' => $name

This associaties a name with the layout for cloning and get_by_name.

=item 'pad' => $pad

This is equivalent to C<($layout = Text::FF::Layout-E<gt>new())-E<gt>pad($pad);>.

=back

=head2 Layout Descriptions

Layout descriptions consist of a set of field specifications. Each field
specification contains a field name, an optional field type, and a position
or size specifier. The layout may be passed as a reference to an array of
tokens, or as a string. If passed as a string, the layout will be broken
into tokens on white space.

Position specifiers may be given as a field width (which begin one character
after the end of the previously specified field), a starting position and a
width given as "start+width", or a starting and ending position given as
"start-end".

If the token after the field name is not one of the above, it is used as
either a type name in the type map (if there is a key in the type map
matching the token) or as the field type itself.

Field types are 'l' for left justified, 'r' for right justified, or '%' and
a sprintf-like format. Within a '%' format, '$' is replaced with the field
width, and '<' or '>' cause the sign to be moved to the far-left or
far-right end of the field, respectively.

The field type may optionally begin with a field width and white space.
In this case, the position or size specification in the layout must be
omitted for fields of this type.

