# Text::FF::Layouts - Parse extended layouts in a language-independent syntax
# (Extends Text::FF::Layout package)
#
# Copyright (c) 2005 Brian Katzung, Kappa Computer Solutions, LLC
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# For more information, contact:
#
#	Kappa Computer Solutions, LLC
#	http://www.kappacs.com/
#
# 2005-03-07	Created by Brian Katzung


package Text::FF::Layout;

use Text::FF::Layout;
use IO::File;

my %type_map_index;
my %included_files;

#----------------------------------------------------------------------
# Parse one extended layout (function)

sub parse_extended_layout
{
	my ($lines) = @_;
	my $line;
	my %layout = (
		'default_type' => 'l',
		'layout' => [],
	);

	while (defined($line = shift(@$lines)))
	{
		next if ($line =~ /^#/);	# Skip comments
		last if ($line =~ /^}/);	# End of block
		$line =~ s/\r//;		# Strip carriage returns

		if ($line =~ /^\@clone (.*)/)
		{
			# Clone another layout
			$layout{'clone'} = $1;
			next;
		}

		if ($line =~ /^\@constant \{/)
		{
			# Define constant field values
			$layout{'constant'} = parse_values($lines);
			next;
		}

		if ($line =~ /^\@default (.*)/)
		{
			# Default field type
			$layout{'default_type'} = $1;
			next;
		}

		if ($line =~ /^\@initial \{/)
		{
			# Define initial field values
			$layout{'initial'} = parse_values($lines);
			next;
		}

		if ($line =~ /^\@map (.*)/)
		{
			# Specify type map to use
			$layout{'type_map'} = $type_map_index{$1};
			next;
		}

		if ($line =~ /^\@pad (.*)/)
		{
			# Enable or disable padding
			$layout{'pad'} = $1;
			next;
		}

		# Add tokens to layout
		push(@{$layout{'layout'}}, split(/\s+/, $line));
	}

	return \%layout;
}

#----------------------------------------------------------------------
# Parse multiple extended layouts (class method)

sub parse_layouts
{
	my ($class, $layouts) = @_;
	my @lines = split(/\n/, $layouts);
	my $line;

	while (defined($line = shift(@lines)))
	{
		next if ($line =~ /^#/);	# Skip comments
		last if ($line =~ /^}/);	# End of block
		$line =~ s/\r//;		# Strip carriage returns

		if ($line =~ /^\@include (.*)/)
		{
			# Include another file
			my $file = $1;
			if (!$included_files{$file} &&
			  ($fd = IO::File->new($file, 'r')))
			{
				unshift(@lines, <$fd>);
				$fd->close();
				++$included_files{$file};
			}
			next;
		}

		if ($line =~ /^\@layout (.*) \{/)
		{
			# Define a new layout
			my $layout = parse_extended_layout(\@lines);
			$class->new(%$layout);
			next;
		}

		if ($line =~ /^\@map (.*) \{/)
		{
			# Define a new type map
			my $name = $1;
			$type_map_index{$name} = parse_values(\@lines);
			next;
		}
	}
}

#----------------------------------------------------------------------
# Parse a key=value list block

sub parse_values
{
	my ($lines) = @_;
	my %hash;
	my $line;

	while (defined($line = shift(@$lines)))
	{
		next if ($line =~ /^#/);	# Skip comments
		last if ($line =~ /^}/);	# End of block
		$line =~ s/\r//;		# Strip carriage returns

		$hash{$1} = $2 if ($line =~ /^[^=]+=(.*)/);
	}

	return \%hash;
}

1;

__END__

=head1 NAME

Text::FF::Layouts - Parse extended layouts

=head1 SYNOPSIS

  use Text::FF::Layouts;

  Text::FF->parse_layouts('
  # comment
  @include file
 
  @map define_map {
  name=[opt_size ]type
  }
 
  @layout define_name {
  @clone use_name
  field [type] spec
  field size_type
  @constant {
  field=value
  }
  @initial {
  field=value
  }
  @default type
  @pad value
  @map use_map
  }
  ');

