# Copyright (c) 2003, 2005 Brian Katzung, Kappa Computer Solutions, LLC
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# For more information, contact:
#
#	Kappa Computer Solutions, LLC
#	http://www.kappacs.com/
#
# 2005-07-13	Added reset() method.
# 2005-03-05	Initial and constant field value support added by
#		Brian Katzung.
# 2003-10-29	Created by Brian Katzung.
#

package Text::FF::Record;

use strict;
use Carp;

#======================================================================
# Functions

#----------------------------------------------------------------------
# Return record value (instance method)

sub get
{
	my ($self) = @_;
	my $layout = $self->layout();

	if ($layout)
	{
		if ($layout->pad())
		{
			# Pad to maximum size
			my $max = $layout->max_size();

			$self->{'value'} .= ' ' x
			  ($max - length($self->{'value'}));
		}
		else
		{
			# Or trim trailing white space
			$self->{'value'} =~ s/ +$//;
		}
	}

	return $self->{'value'};
}

#----------------------------------------------------------------------
# Return all field name/value pairs (instance method)

sub get_all_fields
{
	my ($self) = @_;
	my $layout = $self->layout();

	return () unless ($layout);

	my @names = keys(%{$layout->get_field_specs()});
	my %field;

	@field{@names} = $self->get_field(@names);
	return \%field;
}

#----------------------------------------------------------------------
# Return field value(s) (instance method)

sub get_field
{
	my ($self, @names) = @_;
	my $layout = $self->layout();
	my $rval = $self->{'value'};
	my @fvals;

	return (wantarray? (): undef) unless ($layout);

	my $rval_len = length($rval);
	my $rspec = $layout->get_field_specs();
	foreach (@names)
	{
		my $fspec;

		if ($fspec = $rspec->{$_})
		{
			my ($start, $len, $type) = @$fspec;
			my $fval = '';

			($fval = substr($rval, $start, $len)) =~ s/^ +| +$//g
			  if ($start < $rval_len);

			if ($type =~ /^%.*</)
			{
				# Fix far-left sign if needed
				$fval =~ s/^([+-])( +)/$2$1/;
			}
			elsif ($type =~ /^%.*>/)
			{
				# Fix far-right sign
				$fval =~ s/^( *)(.*)([ +-])$/$1$3$2/;
			}
			push(@fvals, $fval);
		}
		else
		{
			my $name = $layout->{'name'}? " (\"$layout->{'name'}\")": '';
			carp "No field \"$_\" to get in layout $layout$name (record $self)"
			  if ($self->{'strict'}{'get'});
			push(@fvals, undef);
		}
	}

	return (wantarray? @fvals: $fvals[0]);
}

#----------------------------------------------------------------------
# Set/get record layout (instance method)
# Returns the final layout.

sub layout
{
	my ($self, $layout, $init) = @_;

	# Look up layout instance if passed by name
	$layout = Text::FF::Layout->get_by_name($layout)
	  if (defined($layout) && !ref($layout));

	if ($layout)
	{
		# Set layout and apply initial values, if any
		$self->{'layout'} = $layout;
		$self->set_field((!defined($init) || $init)?
		  $layout->get_initial(): ());
	}

	return $self->{'layout'};
}

#----------------------------------------------------------------------
# Create a new record object (class method)
# Returns the instance.

sub new
{
	my ($class, %opt) = @_;
	my $self = bless
	{
		'strict' =>
		{
			'get' => 1,
			'set' => 1,
		},
		'value' => '',
	}, $class;

	$self->layout($opt{'layout'}) if (defined($opt{'layout'}));
	if (defined($opt{'value'}))
	{
		# Set initial record value
		$self->set($opt{'value'});
	}
	elsif ($opt{'initial'})
	{
		# Set initial field values
		$self->set_field($opt{'initial'});
	}

	return $self;
}

#----------------------------------------------------------------------
# Reset record (instance method)
# Returns the instance.

sub reset
{
	my ($self, $layout) = @_;

	delete $self->{'layout'};
	$self->{'value'} = '';

	$self->layout($layout) if (defined($layout));

	return $self;
}

#----------------------------------------------------------------------
# Set record value (instance method)
# Returns instance.

sub set
{
	my ($self, $value) = @_;

	$value =~ s/[\r\n]+//;		# Trim end-of-line characters
	$self->{'value'} = $value;
	$self->set_field();		# Apply constants, if any...

	return $self;
}

#----------------------------------------------------------------------
# Set field value(s) (instance method)
# Returns the instance.

sub set_field
{
	my ($self, @set) = @_;
	my $layout = $self->layout();

	return $self unless ($layout);

	my $rspec = $layout->get_field_specs();
	my $rval = $self->{'value'};

	# Also apply constants after any field changes.
	@set = ((@set && ref($set[0]))? %{$set[0]}: @set,
	  %{$layout->get_constant()});

	while (@set > 1)
	{
		my $fname = shift(@set);
		my $fval = shift(@set);
		my $fspec = $rspec->{$fname};

		unless ($fspec)
		{
			my $lname = $layout->{'name'}? " (\"$layout->{'name'}\")": '';
			carp "No field \"$fname\" to set in layout $layout$lname (record $self)"
			  if ($self->{'strict'}{'set'});
		}

		next unless ($fspec && defined($fval));

		my ($start, $len, $type) = @$fspec;

		#
		# Extend the record value to the start of the field
		# if it isn't already that long
		#
		$rval .= ' ' x ($start - length($rval));

		if ($type =~ /(%.*)/)
		{
			# Format provided
			my $fmt = $1;
			my $lr = '';

			$fmt =~ s/\$/$len/g;
			if ($fmt =~ /([<>])/)
			{
				# Sign at far left (<) or right (>)
				$lr = $1;
				$fmt =~ s/[<>]//;
			}
			$fval = sprintf($fmt, $fval);
			if ($lr eq '<')
			{
				$fval =~ s/^( +)([+-])/$2$1/;
			}
			elsif ($lr eq '>')
			{
				$fval =~ s/^( *)([ +-])(.*)/$1$3$2/;
			}
		}
		elsif ($type =~ /^r/)
		{
			# Right justified
			$fval = sprintf("%${len}s", $fval);
		}
		# else left justified

		if ((my $delta = $len - length($fval)) < 0)
		{
			# Too long; truncate to field width
			$fval = substr($fval, 0, $len);
		}
		else
		{
			# Too short; pad to field width
			$fval .= ' ' x $delta;
		}

		# Replace field in record
		substr($rval, $start, $len) = $fval;
	}

	$self->{'value'} = $rval;
	return $self;
}

#----------------------------------------------------------------------
# Set get/set strictness (instance method)
# Returns the instance.

sub set_strict
{
	my ($self, %strict) = @_;

	%{$self->{'strict'}} = (%{$self->{'strict'}}, %strict);

	return $self;
}

1;

__END__

=head1 NAME

Text::FF::Record - Fixed-field text record class

=head1 SYNOPSIS

  use Text::FF::Record;

  # Create a new record instance
  $record = Text::FF::Record->new(I<options>);

  # Set record layout by name or instance (layout initial values
  # will be applied at this time if $bool is true or absent)
  $record->layout($layout[, $bool]);

  # Get record layout
  $layout = $record->layout();

  # Set/get the record value
  $record->set($value);
  $value = $record->get();

  # Set/get/get all field values
  $record->set_field(%list);    # or { %list }
  @values = $record->get_field(@list);
  %fields = $record->get_all_fields();

  # Reset/clear layout and value
  $record->reset([$layout]);

  # Control error assertion when attempting to get or set
  # non-existent fields
  $record->set_strict('get' => $bool1, 'set' => $bool2);

=head1 DESCRIPTION

This class manages access to fixed-width text fields in records.
The field layouts with the records are controlled by instances of the
Text::FF::Layout class.

=head2 new Options

=item 'initial' => { %list }

Equivalent to C<$record = Text::FF::Record-E<gt>new('layout' =E<gt> $layout)-E<gt>set_field(%list);>.
Requires the layout option. Fields are set from the layout initial values,
then these initial values, then the layout constant values (in that order).

=item 'layout' => $layout

Equivalent to C<$record = Text::FF::Record-E<gt>new()-E<gt>layout($layout);>.

=item 'value' => $value

Equivalent to C<$record = Text::FF::Record-E<gt>new()-E<gt>set($value);>.

